Imports Microsoft.VisualBasic
Imports System
Imports System.ComponentModel
Imports System.ComponentModel.Design
Imports System.Collections
Imports System.Drawing
Imports System.Workflow.ComponentModel.Compiler
Imports System.Workflow.ComponentModel.Serialization
Imports System.Workflow.ComponentModel
Imports System.Workflow.ComponentModel.Design
Imports System.Workflow.Runtime
Imports System.Workflow.Activities
Imports System.Workflow.Activities.Rules

Namespace CustomActivities
	<Designer(GetType(RetryActivityDesigner), GetType(IDesigner))> _
	Partial Public Class RetryActivity
		Inherits CompositeActivity

		Private curRetryIndex As Integer

		<Browsable(False)> _
		Public Property CurrentRetryAttempt() As Integer
			Get
				Return curRetryIndex
			End Get
			Set(ByVal value As Integer)
				curRetryIndex = value
			End Set
		End Property

		Private retryInt As TimeSpan

		Public Property RetryInterval() As TimeSpan
			Get
				Return retryInt
			End Get
			Set(ByVal value As TimeSpan)
				retryInt = value
			End Set
		End Property

		Private SubscriptionID As Guid


		Public Shared RetryCountProperty As DependencyProperty = System.Workflow.ComponentModel.DependencyProperty.Register("RetryCount", GetType(Integer), GetType(RetryActivity), New PropertyMetadata(0))

		<Description("The number of times the execution should be retried"), Category("Retry"), Browsable(True), DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)> _
		Public Property RetryCount() As Integer
			Get
				Return (CInt(Fix(MyBase.GetValue(RetryActivity.RetryCountProperty))))
			End Get
			Set(ByVal value As Integer)
				MyBase.SetValue(RetryActivity.RetryCountProperty, value)
			End Set
		End Property

		Protected Overrides Function Execute(ByVal executionContext As ActivityExecutionContext) As ActivityExecutionStatus
			If EnabledActivities.Count = 0 Then
				Return ActivityExecutionStatus.Closed
			End If

			'try to execute the child activities in a new context
			BeginIteration(executionContext)
			Return ActivityExecutionStatus.Executing
		End Function

		Private Sub BeginIteration(ByVal executionContext As ActivityExecutionContext)
			'create a new activity execution context
			Dim child As Activity = EnabledActivities(0)
			Dim newContext As ActivityExecutionContext = executionContext.ExecutionContextManager.CreateExecutionContext(child)

			'register for both closed and faulted events
			AddHandler newContext.Activity.Closed, AddressOf child_Closed
			AddHandler newContext.Activity.Faulting, AddressOf Activity_Faulting

			'execute the clone activity
			newContext.ExecuteActivity(newContext.Activity)
		End Sub

		Private Sub Activity_Faulting(ByVal sender As Object, ByVal e As ActivityExecutionStatusChangedEventArgs)
			'if the child faults, clear the exception if retrying
			RemoveHandler e.Activity.Faulting, AddressOf Activity_Faulting
			If CurrentRetryAttempt < RetryCount Then
				e.Activity.SetValue(ActivityExecutionContext.CurrentExceptionProperty, Nothing)
			End If
		End Sub

		Private Sub child_Closed(ByVal sender As Object, ByVal e As ActivityExecutionStatusChangedEventArgs)
			Dim thisContext As ActivityExecutionContext = TryCast(sender, ActivityExecutionContext)
			Dim childContext As ActivityExecutionContext = thisContext.ExecutionContextManager.GetExecutionContext(e.Activity)
			RemoveHandler e.Activity.Closed, AddressOf child_Closed

			thisContext.ExecutionContextManager.CompleteExecutionContext(childContext)

			'if the child completed successfully, then we can close
			If e.ExecutionResult = ActivityExecutionResult.Succeeded Then
				Me.SetValue(ActivityExecutionContext.CurrentExceptionProperty, Nothing)
				thisContext.CloseActivity()
				Return
			End If

			'otherwise, if we are not done retrying
			'we need to execute again and make sure to clean up errors
            If CurrentRetryAttempt < RetryCount AndAlso Me.ExecutionStatus = ActivityExecutionStatus.Executing Then
                CurrentRetryAttempt += 1
                Me.SetValue(ActivityExecutionContext.CurrentExceptionProperty, Nothing)

                'setup timer subscription
                Dim expires As DateTime = DateTime.UtcNow.Add(RetryInterval)
                SubscriptionID = Guid.NewGuid()
                Dim qSvc As WorkflowQueuingService = thisContext.GetService(Of WorkflowQueuingService)()
                Dim q As WorkflowQueue = qSvc.CreateWorkflowQueue(SubscriptionID, False)
                AddHandler q.QueueItemAvailable, AddressOf TimerExpired
                Dim subscription As New TimerEventSubscription(SubscriptionID, WorkflowInstanceId, expires)
                Dim timers As TimerEventSubscriptionCollection = GetTimerSubscriptionCollection()
                timers.Add(subscription)

                Return
            Else
                CurrentRetryAttempt += 1
                'if we are done retrying, then we should let the exceptions bubble up
                thisContext.CloseActivity()
            End If
		End Sub

		''' <summary>
		''' Gets the timer subscription collection from the workflow root
		''' </summary>
		''' <returns></returns>
		Private Function GetTimerSubscriptionCollection() As TimerEventSubscriptionCollection
			Dim parent As Activity = Me
			Do While parent.Parent IsNot Nothing
				parent = parent.Parent
			Loop
			Dim timers As TimerEventSubscriptionCollection = CType(parent.GetValue(TimerEventSubscriptionCollection.TimerCollectionProperty), TimerEventSubscriptionCollection)
			Return timers
		End Function


		''' <summary>
		''' Handler that gets notified when the timer has expired
		''' </summary>
		''' <param name="sender"></param>
		''' <param name="e"></param>
		Private Sub TimerExpired(ByVal sender As Object, ByVal e As QueueEventArgs)
			'start a new iteration after the timer expires
			Dim ctx As ActivityExecutionContext = TryCast(sender, ActivityExecutionContext)
			CleanupSubscription(ctx)
			BeginIteration(ctx)
		End Sub

		Protected Overrides Function Cancel(ByVal executionContext As ActivityExecutionContext) As ActivityExecutionStatus
			Dim canClose As Boolean = True
			Dim ctx As ActivityExecutionContext = executionContext

			'cancel any execution iteration
			If ctx.ExecutionContextManager.ExecutionContexts.Count > 0 Then
				ctx.ExecutionContextManager.ExecutionContexts(0).CancelActivity(ctx.ExecutionContextManager.ExecutionContexts(0).Activity)
				canClose = False
			End If

			'remove the timer and related queue if there is a subscription
			CleanupSubscription(ctx)

			Return If(canClose, ActivityExecutionStatus.Closed, ActivityExecutionStatus.Canceling)
		End Function

		''' <summary>
		''' Removes the timersubscription and the workflowqueue if there is one.
		''' </summary>
		''' <param name="ctx"></param>
		Private Sub CleanupSubscription(ByVal ctx As ActivityExecutionContext)
			If SubscriptionID <> Guid.Empty Then
				Dim timers As TimerEventSubscriptionCollection = GetTimerSubscriptionCollection()
				timers.Remove(SubscriptionID)

				ctx.GetService(Of WorkflowQueuingService)().DeleteWorkflowQueue(SubscriptionID)
				SubscriptionID = Guid.Empty
			End If
		End Sub



	End Class
End Namespace
